/*	beep_plugin.c
 *
 *	BEEP-transport plugin for gsoap
 *
 *	Copyright (C) 2003 by Ingo Paschke
 *
 *	Compile & link with gSOAP clients and services to send SOAP
 *	messages over BEEP.
 */

#include <glib.h>
#include <librr/rr.h>
#include "plugin.h"
#include "rrplugin.h"
#ifdef WITH_RULI
#include <ruli.h>
#endif
#define DEBUG(x)

static const char plugin_id[] = PLUGIN_ID;


static int plugin_init (struct soap *soap, struct plugin_data *data);
static int plugin_copy (struct soap *soap, struct soap_plugin *dst,
			struct soap_plugin *src);
static void plugin_delete (struct soap *soap, struct soap_plugin *p);
static int plugin_send (struct soap *soap, const char *buf, size_t len);
static int plugin_fopen (struct soap *soap, const char *endpoint,
			 const char *host, int port);
static int plugin_fpost (struct soap *soap, const char *endpoint,
			 const char *host, int port, const char *path,
			 const char *action, size_t count);
static int plugin_fclose (struct soap *soap);
static int plugin_fpoll (struct soap *soap);
static int plugin_faccept (struct soap *soap, int s, struct sockaddr *a,
			   int *n);
static int plugin_fresponse (struct soap *soap, int soap_error_code,
			     size_t count);
static size_t plugin_recv (struct soap *soap, char *buf, size_t len);

/* plugin: entry point for gsoap: register the plugin */

int
plugin (struct soap *soap, struct soap_plugin *p, void *arg)
{
  p->id = plugin_id;
  p->data = (void *) malloc(sizeof (struct plugin_data));
  p->fcopy = plugin_copy;
  p->fdelete = plugin_delete;
  if (p->data)
    if (plugin_init (soap, (struct plugin_data *) p->data))
    {
      free (p->data);		/* error: could not init */
      return SOAP_EOM;		/* return error */
    }
  return SOAP_OK;
}

/* beep_init
 *
 * convenience function to init the roadrunner-library
 * It's better to use rr_init(&argc, &argv, &error) in main(), as
 * you can't pass runtime-parameters to roadrunner here */

int beep_init(void) {
   GError *error = NULL;
   rr_init(NULL, NULL, &error);
   return 0;
}

/* rr_plugin_start_server
 *
 * Start server: listen for incoming connections and dispatch requests
 * via the soap_serve_callback
 *
 * soap_serve_callback: pointer to the soap_serve(struct soap* soap)
 *                      function
 *
 * hostname, port: address to listen to incoming connections
 */ 

int
rr_plugin_start_server (int (*soap_serve_callback) (struct soap * soap),
			const char *hostname, int port)
{
  RRProfileRegistry *profreg = NULL;
  GError *error = NULL;

  RRListener *listener;

  profreg = rr_profile_registry_new ();
  rr_profile_registry_add_profile (profreg, RR_TYPE_GSOAP, soap_serve_callback);
  listener = rr_tcp_listener_new (profreg, hostname, port, &error);

  if (listener == NULL)
  {
    g_error ("listen failed: %s\n", error->message);
  }

  if (!rr_wait_until_done (&error))
    g_error ("rr_wait failed: %s\n", error->message);

  if (!rr_exit (&error))
    g_error ("rr_exit failed: %s\n", error->message);

  if (!rr_listener_shutdown (listener, &error))
    g_error ("rr_listener_shutdown failed: %s\n", error->message);

  return 0;
}

/* plugin_init: set up plugin_data for gsoap */

static int
plugin_init (struct soap *soap, struct plugin_data *data)
{
  GError *error = NULL;

  data->fsend = soap->fsend;	/* save old recv callback */
  data->frecv = soap->frecv;	/* save old send callback */
  data->fopen = soap->fopen;	/* save old send callback */
  data->fpoll = soap->fpoll;	/* save old send callback */
  data->fpost = soap->fpost;	/* save old send callback */
  data->fclose = soap->fclose;	/* save old send callback */
  data->faccept = soap->faccept;	/* save old send callback */
  data->fresponse = soap->fresponse;	/* save old send callback */
  soap->fsend = plugin_send;	/* replace send callback with ours */
  soap->frecv = plugin_recv;	/* replace recv callback with ours */
  soap->fopen = plugin_fopen;	/* replace recv callback with ours */
  soap->fpoll = plugin_fpoll;	/* replace recv callback with ours */
  soap->fpost = plugin_fpost;	/* replace recv callback with ours */
  soap->fclose = plugin_fclose;	/* replace recv callback with ours */
  soap->faccept = plugin_faccept;	/* replace recv callback with ours */
  soap->fresponse = plugin_fresponse;	/* replace recv callback with ours */
  data->bypass = 0;			/* default: do nothing */
  data->connection = NULL;

  return SOAP_OK;
}

/* plugin_fpoll
 * 
 * beep doesn't use a valid socket, so we don't do anything here
 */

static int plugin_fpoll (struct soap *soap) {
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);
  if (data->bypass) {
    return data->fpoll(soap);
  } else {
    return SOAP_OK;
  }
}

/* plugin_fopen
 *
 * establishes a TCP-connection to endpoint
 *
 * Per RFC3288 we need to do a SRV-DNS-query to determine soap-beep-Servers
 * If hostname and port are known, try to etablish a BEEP-connection. If port
 * is unknown, try a SRV-query and recursively try to connect to the results
 * of that query
 *
 * If endpoint == "peer", try to reuse an established connection
 *
 */

static int
plugin_fopen (struct soap *soap, const char *endpoint, const char *host,
	      int port)
{
  GError *error = NULL;
  int num_servers = 0;
  int i;
  struct _server_list soap_srv_list[MAX_SERVERS];
  int success = 0;

  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);

  /* hack: host == "peer" means connection is alredy established */

  if (strncmp("peer", endpoint, 4) == 0 && data->connection) {
    return SOAP_OK;
  }

  /* only work for soap.beep://-URIs */

  if (strncmp("soap.beep://", endpoint, 7) != 0) {
    data->bypass = 1;
    return data->fopen(soap, endpoint, host, port);
  }

  data->bypass = 0;

  /* we need to get the whole XML-document and enable keepalives */
  
  soap_omode (soap, SOAP_IO_STORE | SOAP_IO_KEEPALIVE);
  soap_imode (soap, SOAP_IO_STORE | SOAP_IO_KEEPALIVE);

  if (port != 0)
  {

    /* we know which port to connect to, so try to establish a connection */

    if (data->connection != NULL) {
      plugin_fclose(soap);
    }

    error = NULL;
    data->profreg = rr_profile_registry_new ();
    rr_profile_registry_add_profile (data->profreg, RR_TYPE_GSOAP, NULL);
    if (error != NULL) {
      g_error("registry_add_profile failed: %s\n", error->message);
      return SOAP_PLUGIN_ERROR;
    }

    if ((data->connection = rr_tcp_connection_new
	 (data->profreg, host, port, &error)) != NULL)
    {
      data->initiator = TRUE;
      return SOAP_OK;
    } else {
      return SOAP_PLUGIN_ERROR;
    }
  }
  else
  {
    /* no port known, get adress and port using SRV-method */
#ifdef WITH_RULI
    num_servers = get_servers (host, &soap_srv_list);
#else
    num_servers = 0;
#endif

    if (num_servers == 0)
    {
      /* SRV failed -> try hostname:10289 */
      return plugin_fopen (soap, endpoint, host, 10289);
    }

    for (i = 0; i < num_servers; i++)
    {
      if (plugin_fopen (soap, endpoint, soap_srv_list[i].host,
			soap_srv_list[i].port) == SOAP_OK)
      {
	success = 1;
	break;
      }
    }
  }
  if (!success)
  {
    g_error ("connection failed: %s\n", error->message);
    return SOAP_EOM;
  }

  /* We are expected to return a valid FD for a socket here, but we override
   * all functions that rely on this */

  return SOAP_OK;
}

/* plugin_new_gsoap
 *
 * called from rrplugin.c for each new connection in server-mode
 * creates a new soap-structure and sets it up
 */

struct soap*
plugin_new_gsoap (RRGSOAP *rrsoap)
{
  struct plugin_data *data;
  struct soap *soap = soap_new();
  
  soap_register_plugin(soap, plugin);

  soap_omode(soap, SOAP_IO_STORE | SOAP_IO_KEEPALIVE);

  data = (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);
  data->rrsoap = rrsoap;
  data->connection = rr_channel_get_connection((RRChannel *)rrsoap);
  soap->keep_alive = 1;
  data->initiator = FALSE;
  return soap;
}

/* plugin_soap_end
 *
 * called from rrplugin.c to free the soap-object used in server-mode
 */

int
plugin_soap_end (struct soap *soap)
{
  soap_end(soap);
  soap_done(soap);
  free(soap);
  return;
}

/* plugin_fpost
 * 
 * This function usually creates the HTTP-Header. We misuse it to start
 * a new channel
 */

static int
plugin_fpost (struct soap *soap, const char *endpoint, const char *host,
	      int port, const char *path, const char *action, size_t count)
{
  GError *error = NULL;
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);

  if (data->bypass) {
    return data->fpost(soap, endpoint, host, port, path, action, count);
  }

  /* If rrsoap is a valid beep-channel, we can safely reuse it here:
   * stdsoap2.c decides if a new connection has to be opened and calls
   * fopen, which closes existing connections and the corresponding
   * channels
   */

  if (RR_IS_GSOAP (data->rrsoap))
  {
    DEBUG (printf ("reusing connection\n"));
    return SOAP_OK;
  }

  /* start a new channel */

  if ((data->rrsoap = rr_gsoap_start (data->connection, host, action,
				      &error)) == NULL)
  {
    if (error)
      g_error ("start request failed: %s\n", error->message);
  }

  g_assert (RR_IS_GSOAP (data->rrsoap));

  return SOAP_OK;
}

/* copy private plugin_data */

static int
plugin_copy (struct soap *soap, struct soap_plugin *dst,
	     struct soap_plugin *src)
{
  struct plugin_data *data;

  /* copy plugin data... */

  /* We don't need to copy the plugin_data struct as the calling code already
   * does this. So we just copy the data private to this plugin: */

  memcpy(dst->data, src->data, sizeof(struct plugin_data));

  /* reset some variables to defaults:
   *
   * Don't reset data->connection (the BEEP TCP-connection) as it may be
   * reused, but reset data->rrsoap (channel-data), so we get to open a new
   * channel for this soap-object */
  
  data = dst->data;
  data->rrsoap = NULL;

  return SOAP_OK;
}

static void
plugin_delete (struct soap *soap, struct soap_plugin *p)
{
  free (p->data); /* free allocated plugin data (this function is not called
		     for shared plugin data) */ 
}

#ifdef WITH_RULI
int
get_servers (const char *host, server_list * soap_srv_list)
{
  ruli_sync_t *sync_query;
  int srv_code, srv_list_size, i;
  ruli_list_t *srv_list;

  /* issue a SRV-query for the given host using the ruli-library */

  sync_query = ruli_sync_query ("_soap-beep._tcp", host, 10289);

  /* Submission failure? */
  if (!sync_query)
  {
    DEBUG (printf ("COULD-NOT-SUBMIT-QUERY\n"));
    return 0;
  }

  srv_code = ruli_sync_srv_code (sync_query);

  /* Service provided ? */
  if (srv_code == RULI_SRV_CODE_UNAVAILABLE)
  {
    DEBUG (printf ("SRV-SERVICE-NOT-PROVIDED-BY-DOMAIN\n"));
    return 0;
  }

  /* Server sent RCODE error? */
  if (srv_code)
  {
    int rcode = ruli_sync_rcode (sync_query);

    if (rcode)
      DEBUG (printf ("BAD-RCODE-IN-SERVER-ANSWER\n"));
    else
      DEBUG (printf ("OTHER-FAILURE\n"));
    return 0;
  }

  /* No error here */

  srv_list = ruli_sync_srv_list (sync_query);
  srv_list_size = ruli_list_size (srv_list);
  srv_list_size = (srv_list_size > MAX_SERVERS) ? MAX_SERVERS : srv_list_size;

  if (srv_list_size < 1)
  {
    DEBUG (printf ("RESULT-ARRAY-IS-EMPTY\n"));
    return 0;
  }

  for (i = 0; i < srv_list_size; ++i)
  {
    ruli_srv_entry_t *entry = ruli_list_get (srv_list, i);
    ruli_list_t *addr_list = &entry->addr_list;
    int addr_list_size = ruli_list_size (addr_list);
    int j;

    {
      char txt_dname_buf[RULI_LIMIT_DNAME_TEXT];
      int txt_dname_len;

      if (ruli_decode_dname (txt_dname_buf, RULI_LIMIT_DNAME_TEXT,
			     &txt_dname_len,
			     entry->target, entry->target_len))
      {

	DEBUG (printf ("target-decoding-failed\n"));
	continue;
      }
      strncpy (soap_srv_list[i].host, txt_dname_buf, HOSTNAMELEN);
      DEBUG (printf ("target=%s ", txt_dname_buf));
    }

    /*
     * Show port number and addresses
     */

    soap_srv_list[i].port = entry->port;
    DEBUG (printf ("port=%d addresses=", entry->port));
  }

  return srv_list_size;
}
#endif

/* plugin_send
 *
 * this is called to send len bytes from buffer buf
 */

static int
plugin_send (struct soap *soap, const char *buf, size_t len)
{
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);

  DEBUG (printf ("\n=send=\n%s\n", buf));

  if (data->bypass)
    return data->fsend(soap, buf, len);

  return rr_gsoap_send (data->rrsoap, buf, len);
}

/* plugin_recv
 *
 * receive len bytes into buffer buf
 */

static size_t
plugin_recv (struct soap *soap, char *buf, size_t len)
{
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);

  if (data->bypass)
    return data->frecv(soap, buf, len);

  size_t res = rr_gsoap_recv (data->rrsoap, buf, len);
  DEBUG (printf ("\n=recv=\n\n", buf));
  return res;
}

/* plugin_fclose
 *
 * close all channels and the tcp-connection
 */

static int
plugin_fclose (struct soap *soap)
{
  GError *error = NULL;
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);

  if (data->bypass) 
    return data->fclose(soap);

  /* only close the connection if this side initiated it, otherwise
   * both sides will attempt to close it */

  if (data->connection != NULL && data->initiator)
  {
    rr_gsoap_close (data->rrsoap, &error);
    g_object_unref(data->connection);
    data->rrsoap = NULL;
    data->connection = NULL;
    if (error)
    {
      g_error("close failed: %s\n", error->message);
      return SOAP_PLUGIN_ERROR;
    }
  }
  return SOAP_OK;
}

/* plugin_fresponse
 * 
 * generates the response-header. We don't need this for soap-beep.
 */

static int
plugin_fresponse (struct soap *soap, int soap_error_code, size_t count)
{
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);

  if (data->bypass)
    return data->fresponse(soap, soap_error_code, count);
 
  return SOAP_OK;
}

/* plugin_faccept
 *
 * accept incoming tcp-connection
 * for beep, use rr_plugin_start_server
 */

static int
plugin_faccept (struct soap *soap, int s, struct sockaddr *a, int *n)
{
  struct plugin_data *data =
    (struct plugin_data *) soap_lookup_plugin (soap, plugin_id);
  if (data->bypass)
    return data->faccept(soap, s, a, n);
  return SOAP_OK;
}
