/*
 * rrplugin.c
 *
 * SOAP/BEEP-plugin for gsoap and roadrunner
 * this handles the interface to the roadrunner library
 *
 * (C) 2003 Ingo Paschke
 *
 * portions taken from Jonas Borgstrm's soap-beep implementation
 * found in the roadrunner cvs repository
 *
 * Copyright (c) 2002 Jonas Borgstrm <jonas@codefactory.se>
 * Copyright (c) 2002, 2003 Daniel Lundin   <daniel@codefactory.se>
 * Copyright (c) 2002 CodeFactory AB.  All rights reserved.
 *
 */

#include <librr/rr.h>
#include "rrplugin.h"
#include "plugin.h"
#include <glib.h>
#include <libxml/tree.h>
#define GSOAP_URI "http://iana.org/beep/soap"
#define SOAP_WINDOW_SIZE (64 * 1024)
#define DEBUG(x)
#define DEBUGA(x)

static GObjectClass *parent_class = NULL;

static gboolean frame_available (RRChannel * channel, RRFrame * frame,
				 GError ** error);

static gboolean client_init (RRChannel * channel, GError ** error);

static gboolean server_init (RRChannel * channel, const gchar * piggyback,
			     GError ** error);

static void close_confirmation (RRChannel * channel, gint code,
				const gchar * xml_lang,
				const gchar * diagnostic);

static void client_confirmation (RRChannel * channel,
				 const gchar * piggyback);

static void finalize (GObject * object);

static gboolean validate_bootmsg (RRGSOAP * soap, const gchar * piggyback,
				  GError ** error);

typedef enum
{
  RR_SOAP_STATE_BOOT = 0,
  RR_SOAP_STATE_READY = 1
} RRSOAPState;

typedef enum
{
  RR_GSOAP_ROLE_SERVER = 0,
  RR_GSOAP_ROLE_CLIENT = 1
} RRGSOAPRole;

struct _RRGSOAPPriv
{
  RRSOAPState state;		/* profile state: boot/ready */
  RRGSOAPRole role;             /* transaction role: client/server */
  GMutex *transaction_mutex;    /* transaction mutex */
  GMutex *active_mutex;         /* mutex to protect private data */
  RRWaiter *waiting;		/* waiter object */
  struct soap *gsoap;		/* gsoap-instance for use in server-mode */
  int (*soap_serve_callback) (struct soap *soap); 
  GError *boot_error;
  int lockcount;		/* DEBUG only */
  int msgno;			/* last BEEP msg-number */
};

/* rr_gsoap_init
 * 
 * initialize instance data
 */

static void
rr_gsoap_init (GObject * object)
{
  RRGSOAP *soap = (RRGSOAP *) object;
  RRGSOAPPriv *priv;
  RRChannel *channel = RR_CHANNEL (object);

  priv = g_new0 (RRGSOAPPriv, 1);
  soap->priv = priv;
  priv->waiting = NULL;
  priv->soap_serve_callback = NULL;
  priv->transaction_mutex = g_mutex_new ();
  priv->active_mutex = g_mutex_new ();
  priv->role = RR_GSOAP_ROLE_CLIENT;
  priv->gsoap = NULL;
  priv->lockcount = 0;
  DEBUG (printf ("rr_gsoap_init\n"));
  rr_channel_set_window_size (channel, SOAP_WINDOW_SIZE);
}

/* rr_gsoap_class_init
 * 
 * initialize class data
 */

static void
rr_gsoap_class_init (GObjectClass * klass)
{
  RRChannelClass *channel_class = (RRChannelClass *) klass;

  channel_class->frame_available = frame_available;
  channel_class->client_init = client_init;
  channel_class->server_init = server_init;
  channel_class->close_confirmation = close_confirmation;
  channel_class->client_confirmation = client_confirmation;

  klass->finalize = finalize;

  parent_class = g_type_class_peek_parent (klass);
  DEBUG (printf ("rr_gsoap_class_init\n"));
}

/* rr_rrsoap_get_type
 *
 * returns GObject type-structure
 */

GType
rr_gsoap_get_type (void)
{
  static GType rr_type = 0;

  if (!rr_type)
  {
    static GTypeInfo type_info = {
      sizeof (RRGSOAPClass),
      NULL,
      NULL,
      (GClassInitFunc) rr_gsoap_class_init,
      NULL,
      NULL,
      sizeof (RRGSOAP),
      16,
      (GInstanceInitFunc) rr_gsoap_init
    };
    rr_type = g_type_register_static (RR_TYPE_CHANNEL, "RRGSOAP",
				      &type_info, 0);
    rr_channel_set_uri (rr_type, GSOAP_URI);
  }
  return rr_type;
}

GQuark
rr_gsoap_error_quark (void)
{
  static GQuark quark;
  if (!quark)
    quark = g_quark_from_static_string ("rr_soap_error");
  return quark;
}

/* finalize
 * 
 * Object destructor
 */

static void
finalize (GObject * object)
{

  RRGSOAP *soap = (RRGSOAP *) object;
  RRGSOAPPriv *priv = soap->priv;

  /* release locks, clean up */

  g_mutex_lock(priv->active_mutex);
  if (priv->gsoap != NULL) {
    plugin_soap_end(priv->gsoap);
  }
  parent_class->finalize (object);
  g_mutex_unlock(priv->active_mutex);
  g_mutex_free (priv->transaction_mutex);
  g_mutex_free (priv->active_mutex);
  g_free (priv);

  DEBUG(printf("end finalize\n"));
}

/* frame_available
 *
 * callback to process available frames
 */

static gboolean
frame_available (RRChannel * channel, RRFrame * frame, GError **error)
{
  RRGSOAP *soap = (RRGSOAP *) channel;
  RRGSOAPPriv *priv = soap->priv;
  gchar *buf;

  if (frame->type == RR_FRAME_TYPE_MSG)
  {
    DEBUG (printf
	   ("------------------->frame received:%d, %d\n", frame->type,
	    priv->state));

    switch (priv->state)
    {
    /* in bootup-phase */
    case RR_SOAP_STATE_BOOT:
      return validate_bootmsg (soap, rr_frame_mime_get_body (frame), error);
      break;
    /* channel established, process messages 
     * frame->type == MSG means that we act as a server here */
    case RR_SOAP_STATE_READY:
      /* start transaction in server-role:
       * expected calls:
       * frecv - get request
       * fsend - send response
       */
      g_mutex_lock (priv->transaction_mutex);
      DEBUG(printf("lock1 %d\n", ++priv->lockcount));
      g_mutex_lock (priv->active_mutex);
      priv->role = RR_GSOAP_ROLE_SERVER;
      /* create new waiter with the received frame */
      priv->waiting = rr_waiter_new ();
      buf = strdup (rr_frame_mime_get_body (frame));
      rr_waiter_set_user_data (priv->waiting, buf);
      DEBUG (printf ("frame avail: %d\n", priv->waiting));
      DEBUG (printf
	     ("waiter: %d:%s", rr_waiter_get_user_data (priv->waiting),
	      rr_waiter_get_user_data (priv->waiting)));
      DEBUG (printf ("recv_buf: %s\n", buf));
      /* this waiter already has its data: */
      rr_waiter_set_done (priv->waiting, NULL);
      /* save message-number so we can set it in the response-packet */
      priv->msgno = frame->msgno;
      g_mutex_unlock (priv->active_mutex);
      /* call gsoap to handle the request */
      priv->soap_serve_callback(priv->gsoap);
      break;
    }
  }
  /* frame->type = RPY means we are the client here
   * Clients are expected to first send the request using fsend,
   * then a reply arrives here which is fetched by frecv, so we're in
   * the middle of a transaction at this point
   */
  else if (frame->type == RR_FRAME_TYPE_RPY)
  {
    g_mutex_lock (priv->active_mutex);
    DEBUG (printf ("Frame type Reply\n"));
    priv->role = RR_GSOAP_ROLE_CLIENT;
    DEBUG (printf ("frame avail: %d\n", priv->waiting));
    buf = strdup (rr_frame_mime_get_body (frame));
    /* set waiter's data */
    rr_waiter_set_user_data (priv->waiting, buf);
    rr_waiter_set_done (priv->waiting, NULL);
    g_mutex_unlock (priv->active_mutex);
  }
  else
  {
    g_set_error (error, RR_BEEP_ERROR, RR_BEEP_CODE_SYNTAX_ERROR,
		 "Unexpected frame received.");
    return FALSE;
  }
  return FALSE;
}

/* rr_gsoap_recv
 *
 * is called to hand len bytes in buf to gsoap
 *
 * TODO: Handle case where len < length(frame)
 */

size_t
rr_gsoap_recv (RRGSOAP * soap, char *buf, size_t len)
{
  GError *error = NULL;
  gchar *response;
  
  RRGSOAPPriv *priv = soap->priv;

  DEBUG (printf ("\n---------------->in soap_recv\n"));

  g_return_val_if_fail (RR_IS_GSOAP (soap), 0);

/* are we waiting for a response? */

  if (priv->waiting == NULL)
  {
    DEBUG (printf ("no waiter\n"));
    g_mutex_unlock (priv->active_mutex);
    return 0;
  }

/* wait for a message */

  DEBUG (printf ("waiting for message\n"));

  if (!rr_waiter_wait (priv->waiting, &error))
  {
    g_mutex_lock(priv->active_mutex);
    rr_waiter_free (priv->waiting);
    priv->waiting = NULL;
    g_mutex_unlock(priv->active_mutex);
    DEBUG(printf("waiter failed: %s\n", error->message));
    return 0;
  }

  /* a frame arrived */

  g_mutex_lock (priv->active_mutex);
  response = rr_waiter_get_user_data (priv->waiting);

  DEBUG (printf ("%s\n", response));
  strncpy (buf, response, len);
  free (response);

  rr_waiter_free (priv->waiting);
  priv->waiting = NULL;

  g_mutex_unlock (priv->active_mutex);

  /* client: done with transaction */
  if (priv->role == RR_GSOAP_ROLE_CLIENT) {
    g_mutex_unlock (priv->transaction_mutex);
    DEBUG(printf("unlock 1 %d\n", --priv->lockcount));
  }
  return strlen (buf);
}

/* rr_gsoap_send
 * sends len bytes from buffer buf on behalf of gsoap-plugin
 */

int
rr_gsoap_send (RRGSOAP * soap, const char *buf, size_t len)
{
  RRGSOAPPriv *priv;
  RRMessage *msg;
  gchar *request;
  GError *error = NULL;
  size_t message_length;
  #define HEADER "Content-Type: application/xml\r\n\r\n"

  g_return_val_if_fail (RR_IS_GSOAP (soap), 0);
  priv = soap->priv;

  /* begin of transaction?
   * Clients begin their transaction here (send->frame_avail->frecv)
   */

  if (priv->role == RR_GSOAP_ROLE_CLIENT) {
  	g_mutex_lock (priv->transaction_mutex);
        DEBUG(printf("lock2 %d\n", ++priv->lockcount));
  }

  DEBUG (printf ("\n------------->in soap_send: %s\n", buf));
  DEBUG (printf ("send: %d\n", priv->waiting));

  g_mutex_lock (priv->active_mutex);
  message_length = strlen(buf);

  /* make sure only len bytes get sent... */
  message_length = (message_length > len) ? len : message_length;

  request = malloc(sizeof(HEADER) + message_length);

  memcpy(request, HEADER, sizeof(HEADER) - 1);
  memcpy(request + sizeof(HEADER) - 1, buf, message_length);
  *(request + sizeof(HEADER) - 1 + message_length) = 0;

  msg = rr_message_static_new ((priv->role == RR_GSOAP_ROLE_SERVER ?
				RR_FRAME_TYPE_RPY : RR_FRAME_TYPE_MSG),
			       (char *) request, strlen (request), TRUE);

  if (priv->role == RR_GSOAP_ROLE_SERVER)
    msg->msgno = priv->msgno;

  if (priv->role == RR_GSOAP_ROLE_CLIENT)
    priv->waiting = rr_waiter_new ();

  if (!rr_channel_send_message (RR_CHANNEL (soap), msg, &error))
  {
    /* cleanup */
    if (priv->role == RR_GSOAP_ROLE_CLIENT)
    {
      rr_waiter_free (priv->waiting);
      priv->waiting = NULL;
    } else {
      /* server: done with transaction */
      g_mutex_unlock (priv->transaction_mutex);
      DEBUG(printf("unlock 2 %d\n", --priv->lockcount));
      priv->role = RR_GSOAP_ROLE_CLIENT;
    }
    g_mutex_unlock (priv->active_mutex);
    return SOAP_PLUGIN_ERROR;
  }
  /* if we're in the server role, the transaction ends here */

  if (priv->role == RR_GSOAP_ROLE_SERVER) {
      g_mutex_unlock (priv->transaction_mutex);
      DEBUG(printf("unlock 3 %d\n", --priv->lockcount));
      priv->role = RR_GSOAP_ROLE_CLIENT;
  }
  g_mutex_unlock (priv->active_mutex);
  return SOAP_OK;
}

/* client_init
 *
 * send bootmessage as client
 */

static gboolean
client_init (RRChannel * channel, GError ** error)
{
  gchar *piggyback;

  if (channel->instance_config)
  {
    piggyback = g_strdup_printf ("<bootmsg resource='%s' />",
				 (gchar *) channel->instance_config);
    rr_channel_set_piggyback (channel, piggyback);
    g_free (piggyback);
  }
  return TRUE;
}

/* validate_bootmsg
 *
 * validate the boot-message when accepting a new beep-connection
 */

static gboolean
validate_bootmsg (RRGSOAP * soap, const gchar * piggyback, GError ** error)
{
  xmlDocPtr doc;
  xmlNodePtr root;
  xmlChar *resource;
  RRGSOAPPriv *priv = soap->priv;

  doc = xmlParseDoc ((xmlChar *) piggyback);
  if (doc == NULL)
  {
    g_set_error (error, RR_BEEP_ERROR, RR_BEEP_CODE_SYNTAX_ERROR,
		 "General syntax error");
    goto fail;
  }

  root = xmlDocGetRootElement (doc);
  if (strcmp (root->name, "bootmsg") != 0)
  {
    g_set_error (error, RR_BEEP_ERROR, RR_BEEP_CODE_PARAM_ERROR,
		 "Syntax error in parameters");
    goto fail;
  }

  resource = xmlGetProp (root, "resource");

  if (resource == NULL)
  {
    g_set_error (error, RR_BEEP_ERROR, RR_BEEP_CODE_PARAM_ERROR,
		 "Syntax error in parameters");
    goto fail;
  }

  /* gsoap doesn't want to know about resources... */

  rr_debug1 ("soap::resource '%s'\n", resource);

  priv->state = RR_SOAP_STATE_READY;
  DEBUG (printf ("validated bootmsg\n"));

  xmlFreeDoc (doc);
  return TRUE;
fail:
  xmlFreeDoc (doc);
  return FALSE;
}

/* server_init
 * 
 * called from the connection_manager when a new channel is opened on this
 * profile
 */


static gboolean
server_init (RRChannel * channel, const gchar * piggyback, GError ** error)
{
  GError *err = NULL;
  RRGSOAP *soap = (RRGSOAP *) channel;
  RRGSOAPPriv *priv = soap->priv;
  priv->role = RR_GSOAP_ROLE_SERVER;

  if (validate_bootmsg (soap, piggyback, &err))
    rr_channel_set_piggyback (channel, "<bootrpy />");
  else
  {
    gchar *str;

    str = g_strdup_printf ("<error code='%d'>%s</error>",
			   err->code, err->message);
    rr_channel_set_piggyback (channel, str);
    g_free (str);
    g_error_free (err);
  }

  /* create new gsoap instance */

  priv->gsoap = (struct soap*)plugin_new_gsoap(soap);
  g_assert(priv->gsoap != NULL);
  priv->soap_serve_callback = channel->global_config;

  return TRUE;
}

static void
close_confirmation (RRChannel * channel, gint code, const gchar * xml_lang,
		    const gchar * diagnostic)
{
  RRGSOAP *soap = (RRGSOAP *) channel;
  RRGSOAPPriv *priv = soap->priv;
  GError *error = NULL;

  g_set_error (&error, RR_GSOAP_ERROR, code, diagnostic);

  g_mutex_lock(priv->active_mutex);
  if (priv->waiting != NULL)
  {
    rr_waiter_set_done (priv->waiting, error);
  }
  g_mutex_unlock(priv->active_mutex);

  g_error_free (error);
}

static void
client_confirmation (RRChannel * channel, const gchar * piggyback)
{
  RRGSOAP *soap = (RRGSOAP *) channel;
  RRGSOAPPriv *priv = soap->priv;

  DEBUG (printf ("in client confirmation\n"));

  if (strstr (piggyback, "<bootrpy"))
  {

    priv->state = RR_SOAP_STATE_READY;
    return;
  }
  else
  {
    xmlDocPtr doc;
    xmlNodePtr node;
    xmlChar *str;
    gint code;

    doc = xmlParseDoc ((xmlChar *) piggyback);
    if (doc == NULL)
      goto err;

    node = xmlDocGetRootElement (doc);

    if (strcmp (node->name, "error") != 0)
      goto err;

    if ((str = xmlGetProp (node, "code")))
    {

      code = atoi (str);
      xmlFree (str);
    }
    if (node->children && node->children->type == XML_TEXT_NODE)
    {
      priv->boot_error = g_error_new (RR_BEEP_ERROR,
				      code, node->children->content);
      DEBUG (printf ("client confirmation error\n"));
      goto out;
    }
  err:
    priv->boot_error = g_error_new (RR_BEEP_ERROR,
				    RR_BEEP_CODE_SYNTAX_ERROR,
				    "blob element parse error.");
  out:
    xmlFreeDoc (doc);
  }
}

RRGSOAP *
rr_gsoap_start (RRConnection * connection, const gchar * server_name,
		const gchar * resource, GError ** error)
{

  RRGSOAP *soap;

  g_return_val_if_fail (RR_IS_CONNECTION (connection), NULL);

  if ((soap = (RRGSOAP *) rr_connection_start (connection,
					       server_name,
					       RR_TYPE_GSOAP,
					       (gchar *) resource,
					       error)) == NULL)
    return NULL;

  if (soap->priv->boot_error)
  {
    g_propagate_error (error, soap->priv->boot_error);
    soap->priv->boot_error = NULL;
    rr_gsoap_close (soap, NULL);
    return NULL;
  }

  soap->priv->role = RR_GSOAP_ROLE_CLIENT;

  return soap;
}

gboolean
rr_gsoap_close (RRGSOAP * soap, GError ** error)
{
  g_return_val_if_fail (RR_IS_GSOAP (soap), FALSE);
  RRGSOAPPriv *priv = soap->priv;

  if (!rr_channel_close (RR_CHANNEL (soap), RR_BEEP_CODE_SUCCESS,
			 NULL, "bye", error))
    return FALSE;

  g_object_unref (G_OBJECT (soap));

  return TRUE;
}
